#ifndef MIXSNET_BERNOUILLI3DDISTRIBUTION_H
#define MIXSNET_BERNOUILLI3DDISTRIBUTION_H
#include<Matrix.hpp>
#include<Algorithms.hpp>
#include<Triplet.hpp>
#include<ColoredWeightedGraph.hpp>
#include<math.h>
#include<limits>
#include<boost/numeric/ublas/io.hpp>
#include<boost/numeric/ublas/matrix.hpp>
#include<boost/random.hpp>
#include <boost/random/uniform_real.hpp>
#include <typeinfo>

namespace mixsnet{
  /*
    Binary coding from right to left :
    000  0
    100  4
    010  2
    110  6
    001  1
    101  5
    011  3
    111  7
   */

  typedef std::array<double, 3> ThreeParams; // 3 x (1 Bernouilli parameter)
  typedef std::array<double, 8> EightParams; // 1 x (8 Bernouilli 3D parameters)

  class Bernouilli3D
  {
    protected:
      color_type _nbcol;
      boost::mt19937 _rng;
      boost::numeric::ublas::matrix<unsigned int> _Nql;
      boost::numeric::ublas::matrix<ThreeParams> _Pql;
      boost::numeric::ublas::matrix<ThreeParams> _Sigmaql;
      boost::numeric::ublas::matrix<double> _Omegaql;
      boost::numeric::ublas::matrix<EightParams> _logP3Dql; // 1 x log(8 Bernouilli 3D parameters)
      void normalizeSigmaqlOmegaql();
      void kroneckerProduct();
      friend class UpdateNqlPqlFunctor;
      template<class TGraph>
      void computeNqlPql(const TGraph& net, const Matrix<Triplet>& Y);
      friend class UpdateSigmaqlOmegaFunctor;
      template<class TGraph>
      void computeSigmaqlOmega(const TGraph& net, const Matrix<Triplet>& Y);
      friend std::ostream& operator<<(std::ostream& out, Bernouilli3D& dist);
      friend std::istream& operator>>(std::istream& in, Bernouilli3D& dist);
    public:
      Bernouilli3D(color_type nbcol)
        : _nbcol(nbcol), _Nql(nbcol, nbcol, 0), _Pql(nbcol, nbcol, ThreeParams()),
          _Sigmaql(nbcol, nbcol, ThreeParams()), _Omegaql(nbcol, nbcol, 0.),
          _logP3Dql(nbcol, nbcol, EightParams()) {
#ifdef _OPENMP
        _rng.seed(time(NULL) + getpid() + omp_get_thread_num());
#else
        _rng.seed(time(NULL) + getpid());
#endif
      }
      Bernouilli3D(const Bernouilli3D& dist)
        : _nbcol(dist._nbcol), _Nql(dist._Nql), _Pql(dist._Pql),
          _Sigmaql(dist._Sigmaql), _Omegaql(dist._Omegaql), _logP3Dql(dist._logP3Dql){
#ifdef _OPENMP
        _rng.seed(time(NULL) + getpid() + omp_get_thread_num());
#else
        _rng.seed(time(NULL) + getpid());
#endif
      }
      virtual ~Bernouilli3D() {}
      void reinit(unsigned int nbcol) {*this = Bernouilli3D(nbcol);}
      unsigned int nbColors() const {return _nbcol;}
      inline double logdensity(unsigned int i, unsigned int j, color_type q, color_type l, Triplet x){
        return _logP3Dql(q,l)[x.binary2int()];
      }
      Triplet draw(unsigned int i, unsigned int j, color_type q, color_type l);
      template<class TGraph>
      void compute(const TGraph& net, const Matrix<Triplet>& Y);
      double getPkql(unsigned int k, color_type q, color_type l) const {return _Pql(q,l)[k];}
      double getSigmakql(unsigned int k, color_type q, color_type l) const {return _Sigmaql(q,l)[k];}
      double getOmegaql(color_type q, color_type l) const {return _Omegaql(q,l);}
      double getP3Dkql(unsigned int k, color_type q, color_type l) const {return exp(_logP3Dql(q,l)[k]);}
  };

  template<class TGraph>
  void Bernouilli3D::compute(const TGraph& net, const Matrix<Triplet>& Y){
    computeNqlPql(net, Y);
    computeSigmaqlOmega(net, Y);
    normalizeSigmaqlOmegaql();
    kroneckerProduct();
  }

  Triplet Bernouilli3D::draw(unsigned int i, unsigned int j, color_type q, color_type l)
  {
    unsigned int k = 0; double sum = exp(_logP3Dql(q,l)[0]);
    boost::uniform_real<> unif(0.,1.);
    double lrand = boost::variate_generator<boost::mt19937&, boost::uniform_real<> >(this->_rng, unif)();
    while (lrand>sum){
        k++;
        sum+=exp(_logP3Dql(q,l)[k]);
    }
    return Triplet(STRTRIPLETS[k]);
  }

  class UpdateNqlPqlFunctor
  {
  private:
    Bernouilli3D& _dist;
  public:
    UpdateNqlPqlFunctor(Bernouilli3D& dist)
    :  _dist(dist) {}
    ~UpdateNqlPqlFunctor() {}
    void operator() (unsigned int i, unsigned j, color_type ci, color_type cj, Triplet& yij){
      _dist._Nql(ci,cj)++;
      for (unsigned int k=0; k<3; k++)
        if (yij[k])
          _dist._Pql(ci,cj)[k] += 1.;
    }
    void exit() {}
  };

  template<class TGraph>
  void Bernouilli3D::computeNqlPql(const TGraph& net, const Matrix<Triplet>& Y){
    _Nql = boost::numeric::ublas::zero_matrix<unsigned int>(_nbcol, _nbcol);
    _Pql = boost::numeric::ublas::zero_matrix<ThreeParams>(_nbcol, _nbcol);
    auto colors = net.getColorProperty();
    forEachYij<void,
    TGraph,
    Matrix<Triplet>,
    UpdateNqlPqlFunctor,
    decltype(colors) >
    (const_cast<TGraph&>(net), const_cast<Matrix<Triplet>&>(Y), UpdateNqlPqlFunctor(*this), colors, true);
    for (color_type q = 0; q<_nbcol; q++){
        for (color_type l = 0; l<_nbcol; l++){// including l==q
            unsigned int nql = _Nql(q,l);
            if (nql>0){
                for (unsigned int k=0; k<3; k++){
                    _Pql(q,l)[k] /= nql;
                }
            }
        }
    }
  }

  class UpdateSigmaqlOmegaFunctor
  {
  private:
    Bernouilli3D& _dist;
  public:
    UpdateSigmaqlOmegaFunctor(Bernouilli3D& dist) : _dist(dist) {}
    ~UpdateSigmaqlOmegaFunctor() {}
    void operator() (unsigned int i, unsigned j, color_type ci, color_type cj, Triplet& yij){
      double d0 = int(yij[0])-_dist._Pql(ci,cj)[0];
      double d1 = int(yij[1])-_dist._Pql(ci,cj)[1];
      double d2 = int(yij[2])-_dist._Pql(ci,cj)[2];
      _dist._Sigmaql(ci,cj)[0] += d0*d1; //  1,2
      _dist._Sigmaql(ci,cj)[1] += d0*d2; //  1,3
      _dist._Sigmaql(ci,cj)[2] += d1*d2; //  2,3
      _dist._Omegaql(ci,cj) += d0*d1*d2;
    }
    void exit() {}
  };

  template<class TGraph>
  void Bernouilli3D::computeSigmaqlOmega(const TGraph& net, const Matrix<Triplet>& Y)
  {
     _Sigmaql = boost::numeric::ublas::zero_matrix<ThreeParams>(_nbcol, _nbcol);
     _Omegaql = boost::numeric::ublas::zero_matrix<double>(_nbcol, _nbcol);
     auto colors = net.getColorProperty();
     forEachYij<void,
     TGraph,
     Matrix<Triplet>,
     UpdateSigmaqlOmegaFunctor,
     decltype(colors) >
     (const_cast<TGraph&>(net), const_cast<Matrix<Triplet>&>(Y), UpdateSigmaqlOmegaFunctor(*this), colors, true);
  }

  void Bernouilli3D::normalizeSigmaqlOmegaql()
  {
    for (color_type q = 0; q<_nbcol; q++){
        for (color_type l = 0; l<_nbcol; l++){// including l==q
            unsigned int nql = _Nql(q,l);
            if (nql>0){
                for (unsigned int k=0; k<3; k++){
                    _Sigmaql(q,l)[k] /= nql;
                }
                _Omegaql(q,l) /= nql;
            }
        }
    }
  }

  void Bernouilli3D::kroneckerProduct()
  {
    for (color_type q = 0; q<_nbcol; q++){
        for (color_type l = 0; l<_nbcol; l++){// including l==q
            // Kronecker product
            ThreeParams& p = _Pql(q,l);
            ThreeParams& s = _Sigmaql(q,l);
            double tmp;
            tmp = (1-p[0])*(1-p[1])*(1-p[2]) + (1-p[2])*s[0] + (1-p[1])*s[1] + (1-p[0])*s[2] - _Omegaql(q,l);
            if (tmp<=0) tmp = std::numeric_limits<double>::min();
            _logP3Dql(q,l)[0] = log(tmp);
            tmp = (1-p[0])*(1-p[1])*p[2]     + p[2]*s[0] - (1-p[1])*s[1] - (1-p[0])*s[2]     + _Omegaql(q,l);
            if (tmp<=0) tmp = std::numeric_limits<double>::min();
            _logP3Dql(q,l)[1] = log(tmp);
            tmp = (1-p[0])*p[1]*(1-p[2])     - (1-p[2])*s[0] + p[1]*s[1] - (1-p[0])*s[2]     + _Omegaql(q,l);
            if (tmp<=0) tmp = std::numeric_limits<double>::min();
            _logP3Dql(q,l)[2] = log(tmp);
            tmp = (1-p[0])*p[1]*p[2]         - p[2]*s[0] - p[1]*s[1] + (1-p[0])*s[2]         - _Omegaql(q,l);
            if (tmp<=0) tmp = std::numeric_limits<double>::min();
            _logP3Dql(q,l)[3] = log(tmp);
            tmp = p[0]*(1-p[1])*(1-p[2])     - (1-p[2])*s[0] - (1-p[1])*s[1] + p[0]*s[2]     + _Omegaql(q,l);
            if (tmp<=0) tmp = std::numeric_limits<double>::min();
            _logP3Dql(q,l)[4] = log(tmp);
            tmp = p[0]*(1-p[1])*p[2]         - p[2]*s[0] + (1-p[1])*s[1] - p[0]*s[2]         - _Omegaql(q,l);
            if (tmp<=0) tmp = std::numeric_limits<double>::min();
            _logP3Dql(q,l)[5] = log(tmp);
            tmp = p[0]*p[1]*(1-p[2])         + (1-p[2])*s[0] - p[1]*s[1] - p[0]*s[2]         - _Omegaql(q,l);
            if (tmp<=0) tmp = std::numeric_limits<double>::min();
            _logP3Dql(q,l)[6] = log(tmp);
            tmp = p[0]*p[1]*p[2]             + p[2]*s[0] + p[1]*s[1] + p[0]*s[2]             + _Omegaql(q,l);
            if (tmp<=0) tmp = std::numeric_limits<double>::min();
            _logP3Dql(q,l)[7] = log(tmp);
        }
    }
  }


  std::ostream& operator<<(std::ostream& out, Bernouilli3D& dist)
  {
    out<<"# P\n";
    for (color_type q = 0; q<dist._nbcol; q++){
        for (color_type l = 0; l<dist._nbcol; l++){
          for (unsigned int k=0; k<3; k++)
            out<<dist.getPkql(k,q,l)<<" ";
          out<<std::endl;
        }
    }
    out<<"# Sigma\n";
    for (color_type q = 0; q<dist._nbcol; q++){
        for (color_type l = 0; l<dist._nbcol; l++){
          for (unsigned int k=0; k<3; k++)
            out<<dist.getSigmakql(k,q,l)<<" ";
          out<<std::endl;
        }
    }
    out<<"# Omega\n";
    for (color_type q = 0; q<dist._nbcol; q++){
        for (color_type l = 0; l<dist._nbcol; l++)
            out<<dist.getOmegaql(q,l)<<" ";
        out<<std::endl;
    }
    out<<"# P3D {000,001,010,011,100,101,110,111}"<<std::endl;
    for (color_type q = 0; q<dist._nbcol; q++){
        for (color_type l = 0; l<dist._nbcol; l++){
          for (unsigned int k=0; k<8; k++)
            out<<dist.getP3Dkql(k,q,l)<<" ";
          out<<std::endl;
        }
    }
    return out;
  }

  std::istream& operator>>(std::istream& in, Bernouilli3D& dist)
  {
    std::string line;
    double x;
    // # P
    getline(in,line);
    for (color_type q = 0; q<dist._nbcol; q++){
        for (color_type l = 0; l<dist._nbcol; l++){
            getline(in,line); std::istringstream linestream(line);
            for (unsigned int k=0; k<3; k++){
                linestream>>x;
                dist._Pql(q,l)[k] = x;
            }
        }
    }
    // # Sigma
    getline(in,line);
    for (color_type q = 0; q<dist._nbcol; q++){
        for (color_type l = 0; l<dist._nbcol; l++){
            getline(in,line); std::istringstream linestream(line);
            for (unsigned int k=0; k<3; k++){
                linestream>>x;
                dist._Sigmaql(q,l)[k] = x;
            }
        }
    }
    // # Omega
    getline(in,line);
    for (color_type q = 0; q<dist._nbcol; q++){
        getline(in,line); std::istringstream linestream(line);
        for (color_type l = 0; l<dist._nbcol; l++){
            linestream>>x;
            dist._Omegaql(q,l) = x;
        }
    }
    // # P3D ?
    int pos = in.tellg();
    getline(in,line);
    if (line[2] == 'P'){ // # P3D
        for (unsigned int k=0; k<(dist._nbcol*dist._nbcol*8); k++)
          getline(in,line); // skipping
    }
    else{
        in.seekg(pos); // back
        dist.kroneckerProduct();
    }
/*
    std::cerr<<"_Pql "<<dist._Pql<<std::endl;
    std::cerr<<"_Sigmaql "<<dist._Sigmaql<<std::endl;
    std::cerr<<"_Omegaql "<<dist._Omegaql<<std::endl;
    */
    return in;
  }
}
#endif
