#ifndef MIXSNET_HETEROGAUSSIANDISTRIBUTION_H
#define MIXSNET_HETEROGAUSSIANDISTRIBUTION_H
#include<Matrix.hpp>
#include<Algorithms.hpp>
#include<ColoredWeightedGraph.hpp>
#include<math.h>
#include<time.h>
#include<random>
#include<chrono>
#include<boost/numeric/ublas/io.hpp>
#include<boost/numeric/ublas/matrix.hpp>
#include<boost/math/distributions/normal.hpp>
#include <typeinfo>

namespace mixsnet{

  class UpdateNqlMuqlFunctor;
  class UpdateSigma2qlFunctor;
  class HeteroscedasticGaussian
  {
  protected:
    color_type _nbcol;
    std::default_random_engine _dre;
    boost::numeric::ublas::matrix<unsigned int> _Nql;
    boost::numeric::ublas::matrix<double> _Muql;
    boost::numeric::ublas::matrix<double> _Sigma2ql;
    virtual void normalizeSigma2ql();
    friend class UpdateNqlMuqlFunctor;
    template<class TGraph, class UpdateFunctor>
    void computeNqlMuql(const TGraph& net, const Matrix<double>& Y);
    friend class UpdateSigma2qlFunctor;
    template<class TGraph, class UpdateFunctor>
    void computeSigma2ql(const TGraph& net, const Matrix<double>& Y);
    friend std::ostream& operator<<(std::ostream& out, HeteroscedasticGaussian& dist);
    friend std::istream& operator>>(std::istream& in, HeteroscedasticGaussian& dist);
  public:
    HeteroscedasticGaussian(color_type nbcol)
    : _nbcol(nbcol), _Nql(nbcol, nbcol, 0), _Muql(nbcol, nbcol, 0.), _Sigma2ql(nbcol, nbcol, 0.) {
#ifdef _OPENMP
      unsigned seed = std::chrono::system_clock::now().time_since_epoch().count();
      _dre.seed(seed);
#else
      unsigned seed = std::chrono::system_clock::now().time_since_epoch().count();
      _dre.seed(seed);
#endif
    }
    HeteroscedasticGaussian(const HeteroscedasticGaussian& dist)
    : _nbcol(dist._nbcol), _Nql(dist._Nql),
      _Muql(dist._Muql), _Sigma2ql(dist._Sigma2ql) {
#ifdef _OPENMP
      unsigned seed = std::chrono::system_clock::now().time_since_epoch().count();
      _dre.seed(seed);
#else
      unsigned seed = std::chrono::system_clock::now().time_since_epoch().count();
      _dre.seed(seed);
#endif
    }
    virtual ~HeteroscedasticGaussian(){}
    void reinit(unsigned int nbcol) {*this = HeteroscedasticGaussian(nbcol);}
    unsigned int nbColors() const {return _nbcol;}
    inline double logdensity(unsigned int i, unsigned int j, color_type q, color_type l,  double x);
    double draw(unsigned int i, unsigned int j, color_type q, color_type l);
    template<class TGraph>
    void compute(const TGraph& net, const Matrix<double>& Y){
      computeNqlMuql<TGraph, UpdateNqlMuqlFunctor>(net, Y);
      //std::cerr<<"_Nql "<<_Nql<<std::endl;
      //std::cerr<<"_Muql "<<_Muql<<std::endl;
      computeSigma2ql<TGraph, UpdateSigma2qlFunctor>(net, Y);
      //std::cerr<<"_Sigma2ql "<<_Sigma2ql<<std::endl;
    }
    double getMuql(color_type q, color_type l) const{
      return _Muql(q,l);
    }
    inline virtual double getSigma2ql(color_type q, color_type l) const{
      if (q<l) std::swap(q, l); // utile??????
      return _Sigma2ql(q,l);
    }
    inline virtual void setSigma2ql(color_type q, color_type l, double sigma2) {_Sigma2ql(q,l) = sigma2;}
    std::default_random_engine& getRandomEngine() {return _dre;}
  };

  double HeteroscedasticGaussian::logdensity(unsigned int i, unsigned int j, color_type q, color_type l, double x)
  {
    if (q<l) std::swap(q, l);
    return log(boost::math::pdf(boost::math::normal(_Muql(q,l), sqrt(this->getSigma2ql(q,l))), x));
    //double mu = _Muql(q,l), sigma = sqrt(this->getSigma2ql(q,l));
    //return exp( -1*(x-mu)*(x-mu) / (2*sigma*sigma)) / (sigma*sqrt(2*3.141592653589793238463));
  }

  double HeteroscedasticGaussian::draw(unsigned int i, unsigned int j, color_type q, color_type l)
  {
    if (q<l) std::swap(q, l);
    std::normal_distribution<> nd(_Muql(q,l), sqrt(this->getSigma2ql(q,l)));
    return nd(_dre);
  }

  class UpdateNqlMuqlFunctor
  {
  private:
    HeteroscedasticGaussian & _dist;
  public:
    UpdateNqlMuqlFunctor(HeteroscedasticGaussian& dist)
    :  _dist(dist) {}
    ~UpdateNqlMuqlFunctor() {}
    void operator() (unsigned int i, unsigned j, color_type ci, color_type cj, double yij){
      if (ci<cj) std::swap(ci, cj);
      _dist._Nql(ci, cj)++;
      _dist._Muql(ci, cj) += yij;
    }
    void exit() {}
  };

  template<class TGraph, class UpdateFunctor>
  void HeteroscedasticGaussian::computeNqlMuql(const TGraph& net, const Matrix<double>& Y)
  {
    _Nql = boost::numeric::ublas::zero_matrix<unsigned int>(_nbcol, _nbcol);
    _Muql = boost::numeric::ublas::zero_matrix<double>(_nbcol, _nbcol);
    auto colors = net.getColorProperty();
    forEachYij<void,
    TGraph,
    Matrix<double>,
    UpdateFunctor,
    decltype(colors) >
    (const_cast<TGraph&>(net), const_cast<Matrix<double>&>(Y), UpdateFunctor(*this), colors);
    for (color_type q = 0; q<_nbcol; q++){
        for (color_type l = 0; l<q; l++){
            unsigned int nql = _Nql(q,l);
            if (nql>0){
                _Muql(q,l) /= nql;
            }
        }
        unsigned int nqq = _Nql(q,q);
        if (nqq>0)
          _Muql(q,q) /= nqq;
    }
  }

  class UpdateSigma2qlFunctor
  {
  private:
    HeteroscedasticGaussian& _dist;
  public:
    UpdateSigma2qlFunctor(HeteroscedasticGaussian& dist) : _dist(dist) {}
    ~UpdateSigma2qlFunctor() {}
    void operator() (unsigned int i, unsigned j, color_type ci, color_type cj, double yij){
      if (ci<cj) std::swap(ci, cj);
      double tmp = yij - _dist._Muql(ci,cj);
      _dist._Sigma2ql(ci, cj) += tmp*tmp;
    }
    void exit() {}
  };

  template<class TGraph, class UpdateFunctor>
  void HeteroscedasticGaussian::computeSigma2ql(const TGraph& net, const Matrix<double>& Y)
  {
    _Sigma2ql = boost::numeric::ublas::zero_matrix<double>(_nbcol, _nbcol);
    auto colors = net.getColorProperty();
    forEachYij<void,
    TGraph,
    Matrix<double>,
    UpdateFunctor,
    decltype(colors) >
    (const_cast<TGraph&>(net), const_cast<Matrix<double>&>(Y), UpdateFunctor(*this), colors);
    normalizeSigma2ql();
  }

  void HeteroscedasticGaussian::normalizeSigma2ql()
  {
    //std::cerr<<"Is it OK?"<<std::endl;
    for (color_type q = 0; q<_nbcol; q++)
      for (color_type l = 0; l<=q; l++){
          unsigned int nql = _Nql(q,l);
          if (nql>0)
            _Sigma2ql(q,l) /= double(nql);
      }
  }

  std::ostream& operator<<(std::ostream& out, HeteroscedasticGaussian& dist)
  {
    out<<"# Mu\n";
    for (color_type q = 0; q<dist._nbcol; q++){
      for (color_type l = 0; l<=q; l++)
        out<<dist._Muql(q,l)<<" ";
      out<<std::endl;
    }
    out<<"# Sigma2\n";
    for (color_type q = 0; q<dist._nbcol; q++){
      for (color_type l = 0; l<=q; l++)
        out<<dist.getSigma2ql(q,l)<<" ";
      out<<std::endl;
    }
    return out;
  }

  std::istream& operator>>(std::istream& in, HeteroscedasticGaussian& dist)
  {
    std::string line;
    double x;
    // # Mu
    getline(in,line);
    for (color_type q = 0; q<dist._nbcol; q++){
        getline(in,line); std::istringstream linestream(line);
        for (color_type l = 0; l<=q; l++){
          linestream>>x;
          dist._Muql(q,l) = x;
        }
        line.clear();
    }
    // # Sigma2
    getline(in,line);
    for (color_type q = 0; q<dist._nbcol; q++){
        getline(in,line); std::istringstream linestream(line);
        for (color_type l = 0; l<=q; l++){
          linestream>>x;
          dist.setSigma2ql(q,l,x);
        }
        line.clear();
    }
    //std::cerr<<"_Nql "<<dist._Nql<<std::endl;
    //std::cerr<<"_Muql "<<dist._Muql<<std::endl;
    //std::cerr<<"_Sigma2ql "<<dist._Sigma2ql<<std::endl;
    return in;
  }

}
#endif
