#include<MixSNet.hpp>
#include<Penalty.hpp>
#include<limits>
#include <math.h>
#ifdef _OPENMP
#include <omp.h>
#endif

namespace mixsnet{

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  class OMPConductor
  {
  private:
    bool _inputallowed;
    bool _hclustallowed;
    bool _spatialallowed;
    bool _randomallowed;
    unsigned int _randomnbtries;
    unsigned int _burnnbiteration;
    double _bestlkl;
    std::vector<color_type> _bestcolors;
    void burnInit(MixSNet<weight_type_t, data_type_t, TDistribution>& model, Matrix<data_type_t>& mat, char initmethod);
  public:
    OMPConductor()
    : _hclustallowed(false), _spatialallowed(false), _randomallowed(false),
    _randomnbtries(0), _burnnbiteration(1), _bestlkl(-std::numeric_limits<double>::max()) {}
    ~OMPConductor() {}
    void activateInput(){_inputallowed = true;}
    void activateHclust(){_hclustallowed = true;}
    void activateSpatial(){_spatialallowed = true;}
    void activateRandom(unsigned int randomnbtries){_randomallowed = true; _randomnbtries = randomnbtries;}
    void setBurnNbIteration(unsigned int burnnbiteration) {_burnnbiteration = burnnbiteration;}
    template<class TPenalty>
    void pipeline(ColoredRoadNetwork<weight_type_t>& net, TDistribution dist, Matrix<data_type_t>& mat, double ldepth, double increment);
  };

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  template<class TPenalty>
  void OMPConductor<weight_type_t, data_type_t, TDistribution>::pipeline(
        ColoredRoadNetwork<weight_type_t>& net,
        TDistribution dist,
        Matrix<data_type_t>& mat,
        double ldepth, double increment)
  {
    MixSNet<weight_type_t, data_type_t, TDistribution> model(net, dist, true);
    if (_inputallowed){
        burnInit(model, mat, 'N');
    }
    if (_hclustallowed){
        burnInit(model, mat, 'H');
    }
    if (_spatialallowed){
        burnInit(model, mat, 'S');
    }
    if (_randomallowed){
        for (unsigned int k=0; k<_randomnbtries; k++){
            burnInit(model, mat, 'R');
        }
    }

    model.initialize(mat, _bestcolors);
    double pen = model.template penaltyTerm<TPenalty>();
    std::cout<<"# Penalty term "<<pen<<std::endl;
    std::cout<<"# Unpenalized Complete likelihodd "<<_bestlkl<<std::endl;
    double lambdamax = 0;
    unsigned int kmax = 1;
    if (pen>0){
      lambdamax = ldepth * abs(_bestlkl) / pen; // empirically lambda*penality comes to 3*completelikelihood
       kmax =int(lambdamax/increment+0.5);
    }
    std::cout<<"LAMBDA\tLKL\tPEN\t";
    for (unsigned int i=0; i<net.nbVertices(); i++) std::cout<<net.getName(net.getVertexByIndex(i))<<" ";
    std::cout<<std::endl;

    std::cout<<"-1 \t"<<_bestlkl<<"\t"<<0<<"\t";
    const std::vector<color_type>& curcolors = model.getCurrentColors();
    for (unsigned int i=0; i<curcolors.size(); i++) std::cout<<curcolors[i]<<" ";
    std::cout<<std::endl;

    for (unsigned int k=0; k<kmax; k++){
        double lambda = k*increment;
        model.setLambda(lambda);
        std::pair<double,double> lklinfo = model.template estimate<TPenalty>(mat);
	std::cout<<lambda<<"\t"<<(lklinfo.first+lklinfo.second)<<"\t"<<lklinfo.second<<"\t";
	const std::vector<color_type>& curcolors = model.getCurrentColors();
	for (unsigned int i=0; i<curcolors.size(); i++) std::cout<<curcolors[i]<<" ";
	std::cout<<std::endl;
	std::cerr<<"### MODEL\n"<<model<<std::endl;
    }
  }

  template<typename weight_type_t, typename data_type_t, class TDistribution>
  void OMPConductor<weight_type_t, data_type_t, TDistribution>::burnInit(MixSNet<weight_type_t, data_type_t, TDistribution>& model, Matrix<data_type_t>& mat, char initmethod)
  {
    model.initialize(mat, initmethod);
    std::pair<double,double> lklinfo = model.template estimate<mixsnet::NoPenalty<weight_type_t> >(mat, _burnnbiteration);
    //std::cerr<<"Lkl "<<initmethod<<" "<<lklinfo.first<<std::endl;
    if (lklinfo.first>_bestlkl){
        _bestlkl = lklinfo.first;
        _bestcolors = model.getCurrentColors();;
    }
  }

}
